unit CoreExports;

interface

uses
  Windows;

procedure LoadDriverLinks;

function GetNumCards: Integer; cdecl;
function QueryCardInfo(CardNum: Integer; pDeviceID, pBase: PLongWord; pNameSize: pLongWord; pName: PAnsiChar): LongWord; cdecl;
function QueryCardID(CardNum: Integer; pDeviceID: PLongWord): LongWord; cdecl;
function QueryCardName(CardNum: Integer; pNameSize: pLongWord; pName: PAnsiChar): LongWord; cdecl;
function QuerySerialNumber(CardNum: Integer; var pSerialNumber: Int64): LongWord; cdecl;
function AIOWDG_CloseDevice(CardNum: Integer): LongWord; cdecl;

function GetDeviceBySerialNumber(const pSerialNumber: Int64): Integer; cdecl;
function GetDeviceByEEPROMByte(Data: Byte): Integer; cdecl;
function GetDeviceByEEPROMData(StartAddress, DataSize: LongWord; pData: PByte): Integer; cdecl;

function CustomEEPROMRead(CardNum: Integer; StartAddress: LongWord; var DataSize: LongWord; Data: Pointer): LongWord; cdecl;
function CustomEEPROMWrite(CardNum: Integer; StartAddress: LongWord; DataSize: LongWord; Data: Pointer): LongWord; cdecl;

function WaitForWDG(CardNum: Integer): LongWord; cdecl;
function AbortWDGRequest(CardNum: Integer): LongWord; cdecl;

function WaitForCOS(CardNum: Integer; pDIData: PByte): LongWord; cdecl;
function WaitForCOSMulti(CardNum: Integer; pTimestamp: PLongInt; DISize: LongInt; pDIData: PByte; pAIData: PDouble): LongWord; cdecl;
function AbortCOSRequest(CardNum: Integer): LongWord; cdecl;

function ReadTimestamp(CardNum: Integer; pTimestamp: PWord): LongWord; cdecl;

function WDGInit(CardNum: Integer): LongWord; cdecl;
function WDGSetTimeout(CardNum: Integer; var pMilliseconds: Double): LongWord; cdecl;
function WDGSetResetDuration(CardNum: Integer; var pMilliseconds: Double): LongWord; cdecl;
function WDGPet(CardNum: Integer): LongWord; cdecl;
function WDGReadTemp(CardNum: Integer; out TempF: Double): LongWord; cdecl;
function WDGReadStatus(CardNum: Integer; out StatusByte: LongWord): LongWord; cdecl;
function WDGStart(CardNum: Integer): LongWord; cdecl;
function WDGStop(CardNum: Integer): LongWord; cdecl;
function WDGHandleIRQ(CardNum: Integer; Action: LongWord): LongWord; cdecl;
function EmergencyReboot(): LongWord; cdecl;

function DI_ReadAll(CardNum: Integer; pDIData: PByte): LongWord; cdecl;
function DI_Read1(CardNum: Integer; BitIndex: Byte; pBitValue: PLongBool): LongWord; cdecl;
function DI_ReadLatched(CardNum: Integer; pDIRisen, pDIFallen: PByte): LongWord; cdecl;
function DI_PeekAll(CardNum: Integer; pDIData: PByte): LongWord; cdecl;
function DI_Peek1(CardNum: Integer; BitIndex: Byte; pBitValue: PLongBool): LongWord; cdecl;
function DI_SetDebounceAll(CardNum: Integer; TimeMS: Byte): LongWord; cdecl;
function DI_SetDebounce1(CardNum: Integer; BitIndex: Byte; TimeMS: Byte): LongWord; cdecl;
function DI_SetDebounceSome(CardNum: Integer; BitMask: Byte; TimeMS: Byte): LongWord; cdecl;
function DO_WriteAll(CardNum: Integer; DOData: Byte): LongWord; cdecl;
function DO_Write1(CardNum: Integer; BitIndex: Byte; BitValue: LongBool): LongWord; cdecl;

function AI_ReadAll(CardNum: Integer; pAIData: PDouble): LongWord; cdecl;
function AI_Read1(CardNum: Integer; ChannelIndex: Byte; pAIData: PDouble): LongWord; cdecl;

function InPortB(Addr: LongWord): Word; cdecl;
function OutPortB(Addr: LongWord; Value: Byte): Word; cdecl;
function RelInPortB(CardNum, Offset: LongWord): Word; cdecl;
function RelOutPortB(CardNum, Offset: LongWord; Value: Byte): Word; cdecl;



function GenericVendorRead(CardNum: Integer; Request: Byte; Value, Index: Word; var DataSize: LongWord; pData: Pointer): LongWord; cdecl;
function GenericVendorWrite(CardNum: Integer; Request: Byte; Value, Index: Word; DataSize: LongWord; pData: Pointer): LongWord; cdecl;
function AWU_GenericBulkIn(CardNum: Integer; PipeID: LongWord; pData: Pointer; DataSize: LongWord; var BytesRead: LongWord): LongWord; cdecl;
function AWU_GenericBulkOut(CardNum: Integer; PipeID: LongWord; pData: Pointer; DataSize: LongWord; var BytesWritten: LongWord): LongWord; cdecl;

function AIOWDG_ResetChip(CardNum: Integer): LongWord; cdecl;





implementation

uses
  Math,
  SysUtils,
  Classes,
  SetupDi,
  //WinUSB,
  ContBufThreads,
  TheseUtils;



const
  ERROR_DEVICE_REMOVED = 1617;
  ERROR_DEVICE_NOT_CONNECTED = 1167; //Observed on an actual motherboard, probably typoed from ERROR_DEVICE_REMOVED.

  FILE_DEVICE_UNKNOWN = $0022;
  METHOD_BUFFERED = 0;
  METHOD_NEITHER  = 3;
  FILE_ANY_ACCESS = 0;

  IOCTL_ADAPT_INDEX = $0000;
  IOCTL_CY_BASE = LongWord(FILE_DEVICE_UNKNOWN shl 16) or (FILE_ANY_ACCESS shl 14);
  IOCTL_ADAPT_SEND_EP0_CONTROL_TRANSFER = IOCTL_CY_BASE or ((IOCTL_ADAPT_INDEX+8) shl 2) or METHOD_BUFFERED;
  IOCTL_ADAPT_SELECT_INTERFACE          = IOCTL_CY_BASE or ((IOCTL_ADAPT_INDEX+3) shl 2) or METHOD_BUFFERED;
  IOCTL_ADAPT_SEND_NON_EP0_DIRECT       = IOCTL_CY_BASE or ((IOCTL_ADAPT_INDEX+18) shl 2) or METHOD_NEITHER;

  AIOWinUSBClassGUID: TGUID = (D1: $88D87F5A; D2: $EA16; D3: $93FC; D4: ($21, $65, $39, $C9, $1C, $36, $C9, $6E) );
  AIOCyUSBClassGUID:  TGUID = (D1: $D52D27E3; D2: $CF76; D3: $25EB; D4: ($72, $EA, $59, $6F, $22, $34, $C0, $E1) );
  D15LoClassGUID:     TGUID = (D1: $AE18AA60; D2: $7F6A; D3: $11D4; D4: ($97, $DD, $00, $01, $02, $29, $B9, $59) );

  CyBaseTimeout = 42; //It's an integer, and all the .PDF says is it's the value you'd expect divided by 1000.

type
  TCySetupPacket = packed record
    bRequestType: Byte;
    bRequest: Byte;
    Value,
    Index,
    Len: Word;
    ulTimeOut: LongWord;
  end;

  TCySingleTransfer = packed record
    SetupPacket: TCySetupPacket;

    reserved: Byte;

    ucEndpointAddress: Byte;
    NtStatus,
    UsbdStatus,
    IsoPacketOffset,
    IsoPacketLength,
    BufferOffset,
    BufferLength: LongWord;
  end;
  PCySingleTransfer = ^TCySingleTransfer;

  TStringArray = array of AnsiString;

  TByteArray = array of Byte;
  TWordArray = array of Word;

  THardByteArray = array[0..32767] of Byte;
  TPHardByteArray = ^THardByteArray;

  THardWordArray = array[0..16383] of Word;
  TPHardWordArray = ^THardWordArray;

  THardDoubleArray = array[0..4095] of Double;
  TPHardDoubleArray = ^THardDoubleArray;

  TOS = (osDunno, osWin95, osWin98, osNT, osNT5);

function Si7013_TempToCelsius(x: word): double;
begin
    Result := (175.72*x)/65536.0 - 46.85;
end;

function Si7013_HumidityToRH(x: word) : double;
begin
    Result :=  125.0*x/65536.0 - 6.0;
end;

function Si7013_CountsToMilliVolts(x: word) : double;
begin
    Result :=  1.25 * x / 32.767;
end;

procedure UnpackAICounts(pAIData: PDouble; pCtSrc: PWord); forward;



const
  AUR_DIO_WRITE           = $10;
  AUR_DIO_READ            = $11;
  AUR_DIO_CONFIG          = $12;
  AUR_DIO_CONFIG_QUERY    = $13;
  AUR_DIO_LATCH_READ      = $14;
  AUR_DIO_DEB_READ        = $15;
  AUR_DIO_ADVANCED        = $18;
  AUR_DIO_CONFIG_MASKED   = $1F;
  AUR_CTR_READ            = $20;
  AUR_CTR_MODE            = $21;
  AUR_CTR_LOAD            = $22;
  AUR_CTR_MODELOAD        = $23;
  AUR_CTR_SELGATE         = $24;
  AUR_CTR_READALL         = $25;
  AUR_CTR_READLATCHED     = $26;
  AUR_CTR_COS_BULK_GATE2  = $27;
  AUR_CTR_COS_BULK_ABORT  = $29;
  {
  AUR_CTR_PUR_FIRST       = $28; //Not used with device, for index offsetting
  AUR_CTR_PUR_OFRQ        = $28; //Set up to output frequency
  AUR_CTR_PUR_MFRQ        = $2C; //Set up to measure frequency
  AUR_CTR_PUR_EVCT        = $2D; //Set up to count events
  AUR_CTR_PUR_MPUL        = $2E; //Set up to measure pulse width
  }
  AUR_WDG_STATUS          = $2E;
  AUR_DIO_WDG16_DEPREC    = $2F;
  AUR_GEN_CLEAR_FIFO_NEXT = $34;
  AUR_GEN_CLEAR_FIFO      = $35;
  AUR_GEN_CLEAR_FIFO_WAIT = $36;
  AUR_GEN_ABORT_AND_CLEAR = $38;
  AUR_SET_GPIF_MODE       = $39;
  AUR_GEN_REGISTER        = $3C;
  AUR_PCIE_WDG            = $40;
  AUR_WDG                 = $44;
  AUR_OFFLINE_READWRITE   = $50;
  AUR_DACDIO_WRITEALL     = $8C;
  //                        $90; //First index assigned to self-tests.
  AUR_SELF_TEST_1         = $91;
  //                        $9F; //Last index assigned to self-tests.
  AUR_DAC_CONTROL         = $B0;
  AUR_DAC_DATAPTR         = $B1;
  AUR_DAC_DIVISOR         = $B2;
  AUR_DAC_IMMEDIATE       = $B3;
  AUR_GEN_STREAM_STATUS   = $B4;
  AUR_FLASH_READWRITE     = $B5;
  AUR_DAC_RANGE           = $B7;
  AUR_PROBE_CALFEATURE    = $BA;
  //                        $BB;
  //                        $BC;
  AUR_DIO_SETCLOCKS       = $BD;
  AUR_ADC_SET_CONFIG      = $BE;
  AUR_ADC_IMMEDIATE       = $BF;
  AUR_DIO_SPI_WRITE       = $C0;
  AUR_DIO_SPI_READ        = $C1;
  AUR_SET_CUSTOM_CLOCKS   = $C4;
  AUR_ADC_GET_CONFIG      = $D2;
  AUR_DEBUG_DEBUG         = $D4;
  AUR_DEBUG_FLASH_1TO1    = $F0;
  AUR_DEBUG_FLASH_ERASE   = $FC;

  //Cypress FX2 vendor requests.
  CUR_RAM_READ            = $A3;



  //For use with AUR_PCIE_WDG.
  AUV_RESET_CARD = $02;
  AUV_RESET_WDG = $52;
  AUV_NOTIRQ_GO = $49;
  AUV_COS_GO = $4A;
  AUV_NOTIRQ_CLEAR = $69;
  AUV_DISABLE = $44;
  AUV_TIMEOUT_SET = $54;
  AUV_PET = $50;



type
  TBulkPseudoIRQThread = class(TStartableThread)
  public
    CardNum: Integer;
  protected
    procedure Execute; override;
  end;

  TWDGConfig = packed record
    Length: Byte;    // = SizeOf(TWDGConfig);
    Rev: Byte;       // = 1;
    Enabled: ByteBool;
    SecondsBeforeBark: Word;
    WDG_RESET_ON_BARK,
    FET_ON_BARK,
    TTL_ON_BARK,
    BUZZER_ON_BARK,
    RELAY_ON_BARK,
    RED_ON_BARK,
    GRN_ON_BARK,
    BLU_ON_BARK: Byte;
    Null: Byte;      // == 0
  end;

  TWDGStatus = packed record
    //Length: Byte;    // = SizeOf(TWDGStatus);
    Config: TWDGConfig; //The Length field indicates the entire struct+1, and the Null field is taken over as the length of itself plus the remainder of the TWDGStatus.
    Rev: Byte;       // = 1;
    TempCounts: Word;
    OldSchoolStatusByte: Byte; //Likely only a few bits implemented.
    SecondsClock: Word;
    SecondsSinceLastPet: Word;
    Null: Byte;      // == 0
  end;

  TPIDCaps = record
    //DIOBytes, DIOConfigBits: LongWord;
    RootClock: Double;
    bGetName: Boolean;
    bCanWDGInit: Boolean;
  end;

  TDeviceData = record
    bDeviceWasHere, bGotHandle: Boolean;
    hHandleTableMutex: THandle;
    HandleTable: array of record
      ThreadID: LongWord;
      hDevice: THandle;
    end;
    DevPath: AnsiString;
    PID: LongWord;

    PIDCaps: TPIDCaps;

    //LastDIOData: array of Byte;

    bWDGInit: Boolean;
    BulkPseudoIRQThread: TBulkPseudoIRQThread;
    WDGAction: LongWord;

    bAIEnabled: Boolean;
  end;
  PDeviceData = ^TDeviceData;

const
  diNone = -1;
  diFirst = -2;
  diOnly = -3;

var
  OS: TOS = osDunno;
  bStupidXPShutdownIssue: Boolean;
  Dev: array[0..31] of TDeviceData;

function _Cy_GenericRead(hDevice: THandle; RequestType, Request: Byte; Value, Index: Word; var DataSize: LongWord; pData: Pointer): LongWord; register; forward;

function Validate(var CardNum: Integer): LongWord;
var
  I: Integer;
  bGotOne, bOnly: Boolean;

begin
  if (CardNum = diOnly) or (CardNum = diFirst) then begin
    bOnly := CardNum = diOnly;
    bGotOne := False;
    for I := 0 to High(Dev) do begin
      if Dev[I].bGotHandle then begin
        if bGotOne then begin
          Result := ERROR_DUP_NAME;
          Exit;
        end;
        CardNum := I;
        if not bOnly then begin
          Result := ERROR_SUCCESS;
          Exit;
        end;
        bGotOne := True;
      end;
    end;
    if (CardNum >= 0) and (CardNum < Length(Dev)) then
      Result := ERROR_SUCCESS
    else
      Result := ERROR_FILE_NOT_FOUND
    ;
  end
  else if (CardNum < 0) or (CardNum >= Length(Dev)) then begin
    Result := ERROR_INVALID_INDEX;
  end
  else
    Result := ERROR_SUCCESS;
  ;
end;

function AIOWDG_CloseDevice(CardNum: Integer): LongWord; cdecl;
var
  I: Integer;
  L: LongWord;

begin
  Result := Validate(CardNum);
  if Result <> ERROR_SUCCESS then Exit;

  with Dev[CardNum] do begin
    if not bGotHandle then begin
      if bDeviceWasHere then begin
        bDeviceWasHere := False;
        Result := ERROR_SUCCESS;
      end
      else
        Result := ERROR_INVALID_HANDLE
      ;
      Exit;
    end;

    if bAIEnabled then begin
      L := 0;
      GenericVendorWrite(CardNum, $40, $1, $0001, L, nil);
      bAIEnabled := False;
    end;

    try
      bDeviceWasHere := False;

      WaitForSingleObject(hHandleTableMutex, INFINITE);
      for I := 0 to High(HandleTable) do
        CloseHandle(HandleTable[I].hDevice)
      ;
      SetLength(HandleTable, 0);
      ReleaseMutex(hHandleTableMutex);

      Result := ERROR_SUCCESS;
    except
      Result := GetLastError;
      if Result = ERROR_SUCCESS then Result := ERROR_GEN_FAILURE;
    end;
    bGotHandle := False;
  end;
end;

function GetThreadHandleForDevice(CardNum: Integer): THandle;
var
  I: Integer;
  ThreadID: LongWord;

begin
  with Dev[CardNum] do begin
    ThreadID := GetCurrentThreadId;

    WaitForSingleObject(hHandleTableMutex, INFINITE);
      Result := INVALID_HANDLE_VALUE;
      for I := 0 to High(HandleTable) do
        if HandleTable[I].ThreadID = ThreadID then begin
          Result := HandleTable[I].hDevice;
          Break;
        end
      ;
      if Result = INVALID_HANDLE_VALUE then begin
        I := Length(HandleTable);
        SetLength(HandleTable, I+1);
        HandleTable[I].ThreadID := ThreadID;
        HandleTable[I].hDevice := CreateFileA(PAnsiChar(DevPath), GENERIC_WRITE or GENERIC_READ, FILE_SHARE_WRITE or FILE_SHARE_READ, nil, OPEN_EXISTING, FILE_FLAG_OVERLAPPED, 0);
        Result := HandleTable[I].hDevice;
      end;
    ReleaseMutex(hHandleTableMutex);
  end;
end;

procedure AddThreadHandleForDevice(CardNum: Integer; hDevice: THandle);
var
  I: Integer;

begin
  with Dev[CardNum] do begin
    WaitForSingleObject(hHandleTableMutex, INFINITE);
      I := Length(HandleTable);
      SetLength(HandleTable, I+1);
      HandleTable[I].ThreadID := GetCurrentThreadId;
      HandleTable[I].hDevice := hDevice;
    ReleaseMutex(hHandleTableMutex);
  end;
end;

function GetDevicePaths(const ClassGUID: TGUID): TStringArray;
var
  DevInterfaceData: SP_DEVICE_INTERFACE_DATA;
  MemberIndex: Integer;
  hDevInfo: THandle;

  L: LongWord;
  pDetailData: PSP_DEVICE_INTERFACE_DETAIL_DATA;

  Buf: AnsiString;
  I, RL: Integer;

begin
  RL := 0;
  SetLength(Result, RL);

  hDevInfo := SetupDiGetClassDevs(@ClassGUID, nil, 0, DIGCF_PRESENT or DIGCF_DEVICEINTERFACE);
  if hDevInfo = 0 then Exit;

  DevInterfaceData.cbSize := sizeof(DevInterfaceData);
  MemberIndex := 0;
  while SetupDiEnumDeviceInterfaces(hDevInfo, nil, ClassGUID, MemberIndex, DevInterfaceData) do begin
    L := 0;
    SetupDiGetDeviceInterfaceDetail(hDevInfo, @DevInterfaceData, nil, 0, @L, nil);

    GetMem(pDetailData, L);
    if SizeOf(Pointer) = 8 then
      pDetailData.cbSize := 8 //The fixed portion is 4. Dunno where 8 comes from.
    else
      pDetailData.cbSize := 5 //The fixed portion is 4. Dunno where 5 comes from.
    ;

    if not SetupDiGetDeviceInterfaceDetail(hDevInfo, @DevInterfaceData, pDetailData, L, nil, nil) then begin
      FreeMem(pDetailData);
      Inc(MemberIndex);
      Continue;
    end;

    SetString(Buf, PAnsiChar(@pDetailData.DevicePath[0]), L - 4);
    FreeMem(pDetailData);
    I := Pos(AnsiChar(#0), Buf);
    if I <> 0 then SetLength(Buf, I - 1);

    SetLength(Result, RL + 1);
    Result[RL] := Buf;
    Inc(RL);
    Buf := '';

    Inc(MemberIndex);
  end;
  SetupDiDestroyDeviceInfoList(hDevInfo);
end;

{
function AnsiMessageBox(hWnd: THandle; Text, Caption: AnsiString; uType: UINT = MB_OK): Integer;
begin
  Result := MessageBox(hWnd, PChar(Text), PChar(Caption), uType);
  Text := '';
  Caption := '';
end;
}

procedure LoadDriverLinks;
var
  DevicePath: TStringArray;
  ThisPath: AnsiString;
  PI, DI: Integer;
  hFile: THandle;
  //hWinUSB: THandle;
  //SetupData: TUSBSetupPacket;
  //ThisSN: Int64;
  //BytesRead: LongWord;
  bFull: Boolean;
  DeviceInfo: USB_DEV_DSC;
  dwBytesReturned: LongWord;
  Status: LongWord;
  ThisPID: Word;
  ThisPIDCaps: TPIDCaps;

begin
  //May want to alter these to not open handles until GetThreadHandleForDevice.

  //CyUSB links.
  DevicePath := GetDevicePaths(AIOCyUSBClassGUID);
  for PI := 0 to High(DevicePath) do begin
    ThisPath := DevicePath[PI];
    for DI := 0 to High(Dev) do
      if Dev[DI].bGotHandle then
        if Dev[DI].DevPath = ThisPath then begin
          ThisPath := '';
          Break;
        end
    ;
    if Length(ThisPath) = 0 then Continue;

    hFile := CreateFileA(PAnsiChar(ThisPath), GENERIC_WRITE or GENERIC_READ, FILE_SHARE_WRITE or FILE_SHARE_READ, nil, OPEN_EXISTING, FILE_FLAG_OVERLAPPED, 0);
    ThisPath := '';
    if hFile = INVALID_HANDLE_VALUE then hFile := 0;
    if hFile = 0 then Continue;

    dwBytesReturned := SizeOf(DeviceInfo);
    Status := _Cy_GenericRead(hFile, $00, $06, $0100, $0000, dwBytesReturned, @DeviceInfo);
    if dwBytesReturned < SizeOf(DeviceInfo) then Status := ERROR_READ_FAULT;
    if Status <> ERROR_SUCCESS then begin
      CloseHandle(hFile);
      Continue;
    end;
    ThisPID := DeviceInfo.idProduct;

    with ThisPIDCaps do begin
      //DIOBytes := 0;
      //DIOConfigBits := $FFFFFFFF;
      RootClock := 0;
      bGetName := False;
      bCanWDGInit := False;

      case ThisPID of
        $8080, $8081: begin //PCIe-WDG-CSM(A)
          bGetName := True;
          //DIOBytes := 2; //?
          bCanWDGInit := True;
        end;
        else begin //AIOUSB may handle it, but we don't.
          CloseHandle(hFile);
          Continue;
        end;
      end;

      //if DIOConfigBits = $FFFFFFFF then DIOConfigBits := DIOBytes;
    end;

    bFull := True;
    for DI := 0 to High(Dev) do with Dev[DI] do
      if (not bGotHandle) and not bDeviceWasHere then begin
        AddThreadHandleForDevice(DI, hFile);
        bGotHandle := True;
        //hWinUSBFile := 0;
        //bCyUSB := True;
        DevPath := DevicePath[PI];

        PID := ThisPID;
        PIDCaps := ThisPIDCaps;

        //SetLength(LastDIOData, PIDCaps.DIOBytes);

        bFull := False;
        Break;
      end
    ;
    if bFull then begin
      CloseHandle(hFile);
      Break;
    end;
  end;
  SetLength(DevicePath, 0);
end;

function GetMappedFailure(CardNum: Integer): LongWord;
begin
  Result := GetLastError;
  case Result of
    ERROR_INVALID_HANDLE, ERROR_DEVICE_REMOVED, ERROR_DEVICE_NOT_CONNECTED, ERROR_FILE_NOT_FOUND: begin
      AIOWDG_CloseDevice(CardNum);
      Dev[CardNum].bDeviceWasHere := True;

      Result := ERROR_DEVICE_REMOVED;
    end;
    ERROR_SUCCESS: Result := ERROR_GEN_FAILURE;
  end;
end;

function EnsureOpen(const CardNum: Integer): LongWord;
var
  Device: PDeviceData;

begin
  Device := @Dev[CardNum];

  if not Device.bGotHandle then begin
    if Device.bDeviceWasHere then
      Result := ERROR_DEVICE_REMOVED
    else
      Result := ERROR_FILE_NOT_FOUND
    ;
    Exit;
  end;

  Result := ERROR_SUCCESS;
end;

function ValidateAndEnsureOpen(var CardNum: Integer): LongWord;
begin
  Result := Validate(CardNum);
  if Result <> ERROR_SUCCESS then Exit;

  Result := EnsureOpen(CardNum);
end;





function _GenericVendorRead(CardNum: Integer; Request: Byte; Value, Index: Word; var DataSize: LongWord; pData: Pointer): LongWord; cdecl;
var
  Buf: AnsiString;
  pCT: PCySingleTransfer;
  BytesReturned: LongWord;

begin
  try
    SetString(Buf, PAnsiChar(pData), DataSize);
  except
    Result := ERROR_NOACCESS;
    Exit;
  end;

  Buf := StringOfChar(AnsiChar(#0), SizeOf(TCySingleTransfer)) + Buf;
  pCT := PCySingleTransfer(PAnsiChar(Buf));
  pCT.SetupPacket.bRequestType := $C0;
  pCT.SetupPacket.bRequest := Request;
  pCT.SetupPacket.Value := Value;
  pCT.SetupPacket.Index := Index;
  pCT.SetupPacket.Len := DataSize;
  pCT.SetupPacket.ulTimeOut := CyBaseTimeout;
  pCT.ucEndpointAddress := 0; //EP0.
  pCT.IsoPacketLength := 0;
  pCT.BufferOffset := SizeOf(TCySingleTransfer);
  pCT.BufferLength := pCT.SetupPacket.Len;
  BytesReturned := 0;
  if DeviceIoControl(GetThreadHandleForDevice(CardNum), IOCTL_ADAPT_SEND_EP0_CONTROL_TRANSFER, PAnsiChar(Buf), Length(Buf), PAnsiChar(Buf), Length(Buf), BytesReturned, nil) then begin
    Result := ERROR_SUCCESS;
    BytesReturned := Min(BytesReturned, Length(Buf));
    if BytesReturned <= SizeOf(TCySingleTransfer) then
      DataSize := 0
    else
      DataSize := BytesReturned - SizeOf(TCySingleTransfer)
    ;
    try
      Move(Buf[1 + SizeOf(TCySingleTransfer)], pData^, DataSize);
    except
      Result := ERROR_NOACCESS;
    end;
  end
  else begin
    Result := GetLastError;
    if Result = ERROR_SUCCESS then Result := ERROR_GEN_FAILURE;
    Buf := '';
  end;
end;

function _GenericVendorWrite(CardNum: Integer; Request: Byte; Value, Index: Word; DataSize: LongWord; pData: Pointer): LongWord; cdecl;
var
  Buf: AnsiString;
  pCT: PCySingleTransfer;
  BytesReturned: LongWord;

begin
  try
    SetString(Buf, PAnsiChar(pData), DataSize);
  except
    Result := ERROR_NOACCESS;
    Exit;
  end;

  Buf := StringOfChar(AnsiChar(#0), SizeOf(TCySingleTransfer)) + Buf;
  pCT := PCySingleTransfer(PAnsiChar(Buf));
  pCT.SetupPacket.bRequestType := $40;
  pCT.SetupPacket.bRequest := Request;
  pCT.SetupPacket.Value := Value;
  pCT.SetupPacket.Index := Index;
  pCT.SetupPacket.Len := DataSize;
  pCT.SetupPacket.ulTimeOut := CyBaseTimeout;
  pCT.ucEndpointAddress := 0; //EP0.
  pCT.IsoPacketLength := 0;
  pCT.BufferOffset := SizeOf(TCySingleTransfer);
  pCT.BufferLength := pCT.SetupPacket.Len;
  BytesReturned := 0;
  if DeviceIoControl(GetThreadHandleForDevice(CardNum), IOCTL_ADAPT_SEND_EP0_CONTROL_TRANSFER, PAnsiChar(Buf), Length(Buf), PAnsiChar(Buf), Length(Buf), BytesReturned, nil) then
    Result := ERROR_SUCCESS
  else
    Result := GetMappedFailure(CardNum)
  ;
end;

function GenericVendorRead(CardNum: Integer; Request: Byte; Value, Index: Word; var DataSize: LongWord; pData: Pointer): LongWord; cdecl;
begin
  Result := ValidateAndEnsureOpen(CardNum);
  if Result <> ERROR_SUCCESS then Exit;

  Result := _GenericVendorRead(CardNum, Request, Value, Index, DataSize, pData);
end;

function GenericVendorWrite(CardNum: Integer; Request: Byte; Value, Index: Word; DataSize: LongWord; pData: Pointer): LongWord; cdecl;
begin
  Result := ValidateAndEnsureOpen(CardNum);
  if Result <> ERROR_SUCCESS then Exit;

  Result := _GenericVendorWrite(CardNum, Request, Value, Index, DataSize, pData);
end;

function _Cy_GenericRead(hDevice: THandle; RequestType, Request: Byte; Value, Index: Word; var DataSize: LongWord; pData: Pointer): LongWord; register;
var
  Buf: AnsiString;
  pCT: PCySingleTransfer;
  BytesReturned: LongWord;

begin
  try
    SetString(Buf, PAnsiChar(pData), DataSize);
  except
    Result := ERROR_NOACCESS;
    Exit;
  end;

  Buf := StringOfChar(AnsiChar(#0), SizeOf(TCySingleTransfer)) + Buf;
  pCT := PCySingleTransfer(PAnsiChar(Buf));
  pCT.SetupPacket.bRequestType := RequestType or $80;
  pCT.SetupPacket.bRequest := Request;
  pCT.SetupPacket.Value := Value;
  pCT.SetupPacket.Index := Index;
  pCT.SetupPacket.Len := DataSize;
  pCT.SetupPacket.ulTimeOut := CyBaseTimeout;
  pCT.ucEndpointAddress := 0; //EP0.
  pCT.IsoPacketLength := 0;
  pCT.BufferOffset := SizeOf(TCySingleTransfer);
  pCT.BufferLength := pCT.SetupPacket.Len;
  BytesReturned := 0;
  if DeviceIoControl(hDevice, IOCTL_ADAPT_SEND_EP0_CONTROL_TRANSFER, PAnsiChar(Buf), Length(Buf), PAnsiChar(Buf), Length(Buf), BytesReturned, nil) then begin
    Result := ERROR_SUCCESS;
    BytesReturned := Min(BytesReturned, Length(Buf));
    if BytesReturned <= SizeOf(TCySingleTransfer) then
      DataSize := 0
    else
      DataSize := BytesReturned - SizeOf(TCySingleTransfer)
    ;
    try
      Move(Buf[1 + SizeOf(TCySingleTransfer)], pData^, DataSize);
    except
      Result := ERROR_NOACCESS;
    end;
  end
  else begin
    Result := GetLastError;
    if Result = ERROR_SUCCESS then Result := ERROR_GEN_FAILURE;
    Buf := '';
  end;
end;

function _hCy_GenericVendorRead(hDevice: THandle; Request: Byte; Value, Index: Word; var DataSize: LongWord; pData: Pointer): LongWord; cdecl;
var
  Buf: AnsiString;
  pCT: PCySingleTransfer;
  BytesReturned: LongWord;

begin
  try
    SetString(Buf, PAnsiChar(pData), DataSize);
  except
    Result := ERROR_NOACCESS;
    Exit;
  end;

  Buf := StringOfChar(AnsiChar(#0), SizeOf(TCySingleTransfer)) + Buf;
  pCT := PCySingleTransfer(PAnsiChar(Buf));
  pCT.SetupPacket.bRequestType := $C0;
  pCT.SetupPacket.bRequest := Request;
  pCT.SetupPacket.Value := Value;
  pCT.SetupPacket.Index := Index;
  pCT.SetupPacket.Len := DataSize;
  pCT.SetupPacket.ulTimeOut := CyBaseTimeout;
  pCT.ucEndpointAddress := 0; //EP0.
  pCT.IsoPacketLength := 0;
  pCT.BufferOffset := SizeOf(TCySingleTransfer);
  pCT.BufferLength := pCT.SetupPacket.Len;
  BytesReturned := 0;
  if DeviceIoControl(hDevice, IOCTL_ADAPT_SEND_EP0_CONTROL_TRANSFER, PAnsiChar(Buf), Length(Buf), PAnsiChar(Buf), Length(Buf), BytesReturned, nil) then begin
    Result := ERROR_SUCCESS;
    BytesReturned := Min(BytesReturned, Length(Buf));
    if BytesReturned <= SizeOf(TCySingleTransfer) then
      DataSize := 0
    else
      DataSize := BytesReturned - SizeOf(TCySingleTransfer)
    ;
    try
      Move(Buf[1 + SizeOf(TCySingleTransfer)], pData^, DataSize);
    except
      Result := ERROR_NOACCESS;
    end;
  end
  else begin
    Result := GetLastError;
    if Result = ERROR_SUCCESS then Result := ERROR_GEN_FAILURE;
    Buf := '';
  end;
end;

function _hCy_GenericVendorWrite(hDevice: THandle; Request: Byte; Value, Index: Word; DataSize: LongWord; pData: Pointer): LongWord; cdecl;
var
  Buf: AnsiString;
  pCT: PCySingleTransfer;
  BytesReturned: LongWord;

begin
  try
    SetString(Buf, PAnsiChar(pData), DataSize);
  except
    Result := ERROR_NOACCESS;
    Exit;
  end;

  Buf := StringOfChar(AnsiChar(#0), SizeOf(TCySingleTransfer)) + Buf;
  pCT := PCySingleTransfer(PAnsiChar(Buf));
  pCT.SetupPacket.bRequestType := $40;
  pCT.SetupPacket.bRequest := Request;
  pCT.SetupPacket.Value := Value;
  pCT.SetupPacket.Index := Index;
  pCT.SetupPacket.Len := DataSize;
  pCT.SetupPacket.ulTimeOut := CyBaseTimeout;
  pCT.ucEndpointAddress := 0; //EP0.
  pCT.IsoPacketLength := 0;
  pCT.BufferOffset := SizeOf(TCySingleTransfer);
  pCT.BufferLength := pCT.SetupPacket.Len;
  BytesReturned := 0;
  if DeviceIoControl(hDevice, IOCTL_ADAPT_SEND_EP0_CONTROL_TRANSFER, PAnsiChar(Buf), Length(Buf), PAnsiChar(Buf), Length(Buf), BytesReturned, nil) then
    Result := ERROR_SUCCESS
  else begin
    Result := GetLastError;
    if Result = ERROR_SUCCESS then Result := ERROR_GEN_FAILURE;
  end;
end;

function AWU_GenericBulkIn(CardNum: Integer; PipeID: LongWord; pData: Pointer; DataSize: LongWord; var BytesRead: LongWord): LongWord; cdecl;
var
  CT: TCySingleTransfer;

begin
  Result := ValidateAndEnsureOpen(CardNum);
  if Result <> ERROR_SUCCESS then begin
    Exit;
  end;

  FillChar(CT, SizeOf(CT), 0);
  CT.ucEndpointAddress := PipeID;
  if DeviceIoControl(GetThreadHandleForDevice(CardNum), IOCTL_ADAPT_SEND_NON_EP0_DIRECT, @CT, SizeOf(CT), pData, DataSize, BytesRead, nil) then begin
    if BytesRead > DataSize then BytesRead := DataSize;
  end
  else begin
    Result := GetMappedFailure(CardNum);
  end;
end;

function AWU_GenericBulkOut(CardNum: Integer; PipeID: LongWord; pData: Pointer; DataSize: LongWord; var BytesWritten: LongWord): LongWord; cdecl;
var
  CT: TCySingleTransfer;

begin
  Result := ValidateAndEnsureOpen(CardNum);
  if Result <> ERROR_SUCCESS then Exit;

  FillChar(CT, SizeOf(CT), 0);
  CT.ucEndpointAddress := PipeID;
  if DeviceIoControl(GetThreadHandleForDevice(CardNum), IOCTL_ADAPT_SEND_NON_EP0_DIRECT, @CT, SizeOf(CT), pData, DataSize, BytesWritten, nil) then begin
    if BytesWritten > DataSize then BytesWritten := DataSize;
  end
  else
    Result := GetMappedFailure(CardNum)
  ;
end;

function _GenericBulkIn(CardNum: Integer; PipeID: LongWord; pData: Pointer; DataSize: LongWord; var BytesRead: LongWord): LongWord;
var
  CT: TCySingleTransfer;

begin
  FillChar(CT, SizeOf(CT), 0);
  CT.ucEndpointAddress := PipeID;
  if DeviceIoControl(GetThreadHandleForDevice(CardNum), IOCTL_ADAPT_SEND_NON_EP0_DIRECT, @CT, SizeOf(CT), pData, DataSize, BytesRead, nil) then begin
    Result := ERROR_SUCCESS;
    if BytesRead > DataSize then BytesRead := DataSize;
  end
  else
    Result := GetMappedFailure(CardNum)
  ;
end;



function GetNumCards: Integer; cdecl;
var
  I: Integer;

begin
  LoadDriverLinks;

  Result := 0;
  for I := 0 to High(Dev) do
    if EnsureOpen(I) = ERROR_SUCCESS then
      Result := I+1
  ;
end;

function QueryCardInfo(CardNum: Integer; pDeviceID, pBase: PLongWord; pNameSize: pLongWord; pName: PAnsiChar): LongWord; cdecl;
begin
  try
    Result := Validate(CardNum);
    if Result <> ERROR_SUCCESS then Exit;

    if not Dev[CardNum].bGotHandle then LoadDriverLinks;

    Result := EnsureOpen(CardNum);
    if Result <> ERROR_SUCCESS then Exit;

    if pDeviceID <> nil then begin
      Result := QueryCardID(CardNum, pDeviceID);
      if Result <> ERROR_SUCCESS then Exit;
    end;
    if pBase <> nil then pBase^ := $8000;
    if pNameSize <> nil then begin
      Result := QueryCardName(CardNum, pNameSize, pName);
      //if Result <> ERROR_SUCCESS then Exit;
    end;
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;

function QueryCardID(CardNum: Integer; pDeviceID: PLongWord): LongWord; cdecl;
begin
  try
    Result := Validate(CardNum);
    if Result <> ERROR_SUCCESS then Exit;

    if not Dev[CardNum].bGotHandle then LoadDriverLinks;

    Result := EnsureOpen(CardNum);
    if Result <> ERROR_SUCCESS then Exit;

    if pDeviceID = nil then begin
      Result := ERROR_INVALID_ACCESS;
      Exit;
    end;

    pDeviceID^ := Dev[CardNum].PID;
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;

function QueryCardName(CardNum: Integer; pNameSize: pLongWord; pName: PAnsiChar): LongWord; cdecl;
var
  Content: AnsiString;
  wContent: WideString;
  L, L2: LongWord;

begin
  try
    Result := Validate(CardNum);
    if Result <> ERROR_SUCCESS then Exit;

    if not Dev[CardNum].bGotHandle then LoadDriverLinks;

    Result := EnsureOpen(CardNum);
    if Result <> ERROR_SUCCESS then Exit;

    if pNameSize = nil then begin
      Result := ERROR_INVALID_ACCESS;
      Exit;
    end;

    if not Dev[CardNum].PIDCaps.bGetName then begin //Old ASM firmwares do not support get-string
      pNameSize^ := 0;
    end
    else begin
      SetString(wContent, nil, pNameSize^ + 1);
      L := Length(wContent) * SizeOf(wContent[1]);
      Result := GenericVendorRead(CardNum, $06, $0302, 0, L, PWChar(wContent)); //Cypress get-string
      if Result <> ERROR_SUCCESS then Exit;

      L2 := Byte(wContent[1]) div 2;
      if L2 <> 0 then begin
        Dec(L2);
        if pName <> nil then begin
          L := Min(pNameSize^, L2+1);
          wContent := Copy(wContent + #0, 2, L);
          Content := AnsiString(wContent);
          Move(Content[1], pName^, L);
        end;
      end;
      Result := ERROR_SUCCESS;
      pNameSize^ := L2;
    end;
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;

function ClearDevices: LongWord; cdecl;
var
  I: Integer;
  wContent: WideString;
  L: LongWord;
  ErrorCode: LongWord;

begin
  for I := High(Dev) downto 0 do begin
    if Dev[I].bGotHandle then begin
      if not Dev[I].PIDCaps.bGetName then begin
        //Old ASM firmwares do not support get-string
      end
      else begin
        ErrorCode := EnsureOpen(I);
        if ErrorCode = ERROR_SUCCESS then begin
          SetString(wContent, nil, 257);
          L := Length(wContent) * SizeOf(wContent[1]);
          ErrorCode := GenericVendorRead(I, $06, $0302, 0, L, PWChar(wContent));

          if ErrorCode <> ERROR_SUCCESS then
            GetMappedFailure(I)
          ;
        end;
      end;
    end
    else begin
      Dev[I].bDeviceWasHere := False;
    end;
  end;
  Result := ERROR_SUCCESS;
end;

function GetDeviceBySerialNumber(const pSerialNumber: Int64): Integer; cdecl;
var
  I: Integer;
  ThisSerialNumber: Int64;
  Status: LongWord;

begin
  for I := 0 to High(Dev) do begin
    Status := QuerySerialNumber(I, ThisSerialNumber);
    if Status <> ERROR_SUCCESS then Continue;
    if ThisSerialNumber <> pSerialNumber then Continue;

    Result := I;
    Exit;
  end;

  Result := diNone;
end;

function GetDeviceByEEPROMByte(Data: Byte): Integer; cdecl;
var
  ThisData: Byte;
  Status, ThisDataSize, MyLastError: LongWord;
  I: Integer;

begin
  Result := diNone;
  MyLastError := ERROR_FILE_NOT_FOUND;

  for I := 0 to High(Dev) do begin
    ThisDataSize := 1;
    Status := CustomEEPROMRead(I, 0, ThisDataSize, @ThisData);
    if Status <> ERROR_SUCCESS then Continue;
    if ThisDataSize <> 1 then Continue;
    if ThisData <> Data then Continue;

    if Result = diNone then begin
      Result := I;
      MyLastError := ERROR_SUCCESS;
    end
    else begin
      MyLastError := ERROR_DUP_NAME;
    end;
  end;
  SetLastError(MyLastError);
end;

function GetDeviceByEEPROMData(StartAddress, DataSize: LongWord; pData: PByte): Integer; cdecl;
var
  ipData: PByte;
  ThisData: array of Byte;
  Status, ThisDataSize, MyLastError: LongWord;
  I, J: Integer;
  bMatch: Boolean;

begin
  Result := diNone;
  MyLastError := ERROR_FILE_NOT_FOUND;

  if DataSize > $100 then begin
    SetLastError(ERROR_BAD_LENGTH);
    Exit;
  end;

  if DataSize = 0 then begin
    Result := 0;
    if not Dev[Result].bGotHandle then Result := diNone;
    Exit;
  end;

  SetLength(ThisData, DataSize);
  for I := 0 to High(Dev) do begin
    ThisDataSize := DataSize;
    Status := CustomEEPROMRead(I, StartAddress, ThisDataSize, @ThisData[0]);
    if Status <> ERROR_SUCCESS then Continue;
    if ThisDataSize <> DataSize then Continue;

    bMatch := True;
    ipData := pData;
    for J := 0 to DataSize - 1 do begin
      if ThisData[J] <> ipData^ then begin
        bMatch := False;
        Break;
      end;
      Inc(ipData);
    end;
    if not bMatch then Continue;

    if Result = diNone then begin
      Result := I;
      MyLastError := ERROR_SUCCESS;
    end
    else begin
      MyLastError := ERROR_DUP_NAME;
    end;
  end;
  SetLastError(MyLastError);
end;





function CustomEEPROMRead(CardNum: Integer; StartAddress: LongWord; var DataSize: LongWord; Data: Pointer): LongWord; cdecl;
begin
  if StartAddress > $FF then begin
    Result := ERROR_SEEK;
    Exit;
  end;
  if DataSize > (LongWord($100) - StartAddress) then begin
    Result := ERROR_BAD_LENGTH;
    Exit;
  end;

  Result := GenericVendorRead(CardNum, $A2, $1E00 + StartAddress, 0, DataSize, Data);
end;

function CustomEEPROMWrite(CardNum: Integer; StartAddress: LongWord; DataSize: LongWord; Data: Pointer): LongWord; cdecl;
const
  BlockSize = 32;

begin
  if StartAddress > $FF then begin
    Result := ERROR_SEEK;
    Exit;
  end;
  if DataSize > (LongWord($100) - StartAddress) then begin
    Result := ERROR_BAD_LENGTH;
    Exit;
  end;

  while DataSize > BlockSize do begin
    Result := GenericVendorWrite(CardNum, $A2, $1E00 + StartAddress, 0, BlockSize, Data);
    if Result <> ERROR_SUCCESS then Exit;
    Inc(StartAddress, BlockSize);
    Dec(DataSize, BlockSize);
    Inc(PByte(Data), BlockSize);
  end;
  Result := GenericVendorWrite(CardNum, $A2, $1E00 + StartAddress, 0, DataSize, Data);
end;

function GetDeviceHandle(CardNum: Integer; Data: PLongWord): LongWord; cdecl;
begin
  Result := ValidateAndEnsureOpen(CardNum);
  if Result <> ERROR_SUCCESS then Exit;

  Data^ := GetThreadHandleForDevice(CardNum);
end;

function AIOUSB_UploadFirmware(CardNum: Integer; FirmBuf: PAnsiChar; BufLen: LongWord): LongWord; cdecl;
begin
  Result := ERROR_NOT_SUPPORTED;
end;

function StrToLines(S: AnsiString): TStringArray;
var
  Count, Capacity, iC, oC,
  L: Integer;

begin
  Count := 0;
  Capacity := 8;
  SetLength(Result, Capacity);
  S := S + #13;
  L := Length(S);
  iC := 1;
  oC := 1;
  while iC <= L do begin
    if (S[iC] = #13) or (S[iC] = #10) then begin
      if Count >= Capacity then begin
        Capacity := Capacity * 2;
        SetLength(Result, Capacity);
      end;
      Result[Count] := Copy(S, oC, iC - oC);
      Inc(Count);

      if (iC < L) and (S[iC] = #13) and (S[iC+1] = #10) then Inc(iC);
      oC := iC+1;
    end;
    Inc(iC);
  end;
  SetLength(Result, Count);
end;

function AIOUSB_UploadD15LoFirmwaresByPID(pFirmScript: PAnsiChar): LongWord; cdecl;
type
  THexChunk = record
    Addr: Word;
    Data: AnsiString;
  end;
  THexChunks = array of THexChunk;

  TPIDFirmData = record
    PID: Word;
    FirmContent: THexChunks;
  end;

  function ParseHex(HexStrs: TStringArray): THexChunks;
  var
    iLine, iChunk: Integer;
    ThisAddr, ThisData: AnsiString;

  begin
    //Validate hex file.
    SetLength(Result, 0);
    for iLine := 0 to High(HexStrs) do begin
      ThisData := HexStrs[iLine];
      if ThisData = '' then Continue;
      if ThisData[1] <> ':' then Exit;
      if not IsHex(Copy(ThisData, 2, MAXINT)) then Exit;
      if (Length(ThisData) and 1) = 0 then Exit; //If source line is even(has colon plus odd nybbles), reject as invalid.
    end;

    SetLength(Result, Length(HexStrs));
    iChunk := 0;
    for iLine := 0 to High(HexStrs) do begin
      ThisData := HexStrs[iLine];
      if ThisData = '' then Continue;
      if Copy(ThisData, 8, 2) <> '00' then Continue; //We only care about type 00 records.

      ThisAddr := Copy(ThisData, 4, 4);
      ThisData := Copy(ThisData, 10, Length(ThisData) - 11);

      Result[iChunk].Addr := HexToInt(ThisAddr);
      Result[iChunk].Data := HexToStr(ThisData);
      Inc(iChunk);
    end;
    SetLength(Result, iChunk);
  end;

var
  PFD: array of TPIDFirmData; //Loading the script would be a bit faster with pointers, but that's not much of an improvement, so safety is more important.
  iPF, iLine, iPath, iFirmware,
  I: Integer;
  ThisLine, PIDStr, FirmStr: AnsiString;
  DevicePath, FirmScript: TStringArray;
  FirmHex: THexChunks;
  hCyFile: THandle;
  Status: LongWord;
  CPUCSByte: Byte;
  ThisPID: Word;

begin
  Result := 0;
  try
    FirmScript := StrToLines(pFirmScript);
    iPF := -1;
    for iLine := 0 to High(FirmScript) do begin
      ThisLine := FirmScript[iLine];
      if ThisLine = '' then Continue;
      if ThisLine[1] = ';' then Continue;

      SplitStr(ThisLine, '=', PIDStr, FirmStr);
      if PIDStr = '' then Continue;
      if FirmStr = '' then Continue;
      if not IsHex(PIDStr) then Continue;
      if not FileExists(String(FirmStr)) then Continue;

      try
        FirmStr := DumpFileToString(FirmStr);
        FirmStr := AnsiString(UpperCase(Trim(String(FirmStr))));
        FirmHex := ParseHex(StrToLines(FirmStr));
        if Length(FirmHex) = 0 then Continue;
      except
        Continue;
      end;

      Inc(iPF);
      SetLength(PFD, iPF+1);
      with PFD[iPF] do begin
        PID := HexToInt(PIDStr) and $7FFF;
        FirmContent := FirmHex;
      end;
    end;

    DevicePath := GetDevicePaths(D15LoClassGUID);
    for iPath := 0 to High(DevicePath) do begin
      ThisLine := DevicePath[iPath];

      I := Pos('PID_', UpperCase(String(ThisLine)));
      if I = 0 then Continue;
      PIDStr := Copy(ThisLine, I + 4, 4);
      if not IsHex(PIDStr) then Continue;
      ThisPID := HexToInt(PIDStr) and $7FFF;
      iFirmware := High(PFD);
      while (iFirmware >= 0) and (PFD[iFirmware].PID <> ThisPID) do Dec(iFirmware);
      if iFirmware < 0 then Continue; //PID not recognized, don't bother.

      hCyFile := CreateFileA(PAnsiChar(ThisLine), GENERIC_WRITE or GENERIC_READ, FILE_SHARE_WRITE or FILE_SHARE_READ, nil, OPEN_EXISTING, FILE_FLAG_OVERLAPPED, 0);
      ThisLine := '';
      if hCyFile = INVALID_HANDLE_VALUE then hCyFile := 0;
      if hCyFile = 0 then Continue;

      //Transfer the hex file.

      //Put CPU into reset.
      CPUCSByte := $01;
      Status := _hCy_GenericVendorWrite(hCyFile, $A0, $E600, 0, 1, @CPUCSByte);
      if Status = ERROR_SUCCESS then begin
        //Transfer each line.
        for iLine := 0 to High(PFD[iFirmware].FirmContent) do
          with PFD[iFirmware].FirmContent[iLine] do
            _hCy_GenericVendorWrite(hCyFile, $A0, Addr, 0, Length(Data), PAnsiChar(Data))
        ;
        //take CPU out of reset.
        CPUCSByte := $00;
        _hCy_GenericVendorWrite(hCyFile, $A0, $E600, 0, 1, @CPUCSByte);
        Inc(Result);
      end;

      CloseHandle(hCyFile);
    end;
    SetLastError(ERROR_SUCCESS);
  except
    SetLastError(ERROR_INTERNAL_ERROR);
  end;
end;

function QuerySerialNumber(CardNum: Integer; var pSerialNumber: Int64): LongWord; cdecl;
var
  L: LongWord;
  Content: Int64;

begin
  Result := ValidateAndEnsureOpen(CardNum);
  if Result <> ERROR_SUCCESS then Exit;

  L := SizeOf(Content);
  Result := GenericVendorRead(CardNum, $A2, $1DF8, 0, L, @Content);

  if (Result = ERROR_SUCCESS) and (L = SizeOf(Content)) then begin
    if Content = 0 then Content := -1;
    pSerialNumber := Content;
    if Content = -1 then Result := ERROR_NOT_SUPPORTED;
  end
  else begin
    Result := GetMappedFailure(CardNum);
  end;
end;





function WaitForWDG(CardNum: Integer): LongWord; cdecl;
var
  IRQData: array of Byte;
  BytesRead: LongWord;

begin
  SetLength(IRQData, 5);
  ZeroMemory(@IRQData[0], Length(IRQData));
  BytesRead := 0;
  Result := AWU_GenericBulkIn(CardNum, $82, @IRQData[0], Length(IRQData), BytesRead);
  if Result <> ERROR_SUCCESS then Exit;

  if BytesRead <> LongWord(Length(IRQData)) then
    Result := ERROR_INVALID_DATA
  else if (IRQData[4] and 1) = 0 then
    Result := ERROR_OPERATION_ABORTED
  else
    Result := ERROR_SUCCESS
  ;
end;

function AbortWDGRequest(CardNum: Integer): LongWord; cdecl;
var
  IRQData: array of Byte;

begin
  SetLength(IRQData, 5);
  ZeroMemory(@IRQData[0], Length(IRQData));
  Result := GenericVendorWrite(CardNum, AUR_PCIE_WDG, 0, AUV_NOTIRQ_GO, 0, nil);
end;



function _WaitForCOSAll(CardNum: Integer; var ErrorResult: LongWord): TByteArray;
var
  BytesRead: LongWord;

begin
  SetLength(Result, 512);
  ZeroMemory(@Result[0], Length(Result));
  BytesRead := 0;
  ErrorResult := AWU_GenericBulkIn(CardNum, $88, @Result[0], Length(Result), BytesRead);
  SetLength(Result, BytesRead);
  if ErrorResult <> ERROR_SUCCESS then Exit;

  if BytesRead <= 4 then
    ErrorResult := ERROR_INVALID_DATA
  else if (Result[4] and 1) = 0 then
    ErrorResult := ERROR_OPERATION_ABORTED
  else
    ErrorResult := ERROR_SUCCESS
  ;
end;

function WaitForCOS(CardNum: Integer; pDIData: PByte): LongWord; cdecl;
var
  IRQData: TByteArray;

begin
  try
    pDIData^ := 0;
  except
    Result := ERROR_NOACCESS;
    Exit;
  end;

  try
    IRQData := _WaitForCOSAll(CardNum, Result);
    if Result = ERROR_SUCCESS then
      pDIData^ := IRQData[6]
    ;
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;

function WaitForCOSMulti(CardNum: Integer; pTimestamp: PLongInt; DISize: LongInt; pDIData: PByte; pAIData: PDouble): LongWord; cdecl;
var
  IRQData: TByteArray;
  DstOff: Integer;

  procedure CopyDIByte(SrcOff: Integer);
  begin
    Inc(DstOff);

    if DISize <= SrcOff then Exit;

    if Length(IRQData) <= SrcOff then begin
      Result := ERROR_INVALID_DATA;
      TPHardByteArray(pDIData)[DstOff] := 0;
    end
    else begin
      TPHardByteArray(pDIData)[DstOff] := IRQData[SrcOff];
    end;
  end;

begin
  if pTimestamp <> nil then try
    pTimestamp^ := 0;
  except
    Result := ERROR_NOACCESS;
    Exit;
  end;

  if (DISize > 0) and (pDIData <> nil) then try
    ZeroMemory(pDIData, DISize);
  except
    Result := ERROR_NOACCESS;
    Exit;
  end;

  if pAIData <> nil then try
    ZeroMemory(pAIData, 6*2);
  except
    Result := ERROR_NOACCESS;
    Exit;
  end;

  try
    IRQData := _WaitForCOSAll(CardNum, Result);
    if Result <> ERROR_SUCCESS then Exit;
    if Length(IRQData) < 27 then begin
      Result := ERROR_INVALID_DATA;
      Exit;
    end;

    if pTimestamp <> nil then
      pTimestamp^ := PWord(@IRQData[2])^ //USB frame from [2] and [3].
    ;
    if (DISize > 0) and (pDIData <> nil) then begin
      DstOff := -1;
      CopyDIByte( 6); //[0] Debounced DI data.
      CopyDIByte(21); //[1] Debounced COS changes.
      CopyDIByte(24); //[2] COS rising.
      CopyDIByte(26); //[3] COS falling.
      CopyDIByte( 7); //[4] DI rising.
      CopyDIByte( 8); //[5] DI falling.
      CopyDIByte( 5); //[6] Raw DI data.
    end;
    if pAIData <> nil then
      UnpackAICounts(pAIData, @IRQData[9])
    ;
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;

function AbortCOSRequest(CardNum: Integer): LongWord; cdecl;
var
  IRQData: array of Byte;

begin
  SetLength(IRQData, 5);
  ZeroMemory(@IRQData[0], Length(IRQData));
  Result := GenericVendorWrite(CardNum, AUR_PCIE_WDG, 0, AUV_COS_GO, 0, nil);
end;



function ReadTimestamp(CardNum: Integer; pTimestamp: PWord): LongWord; cdecl;
const
  EP0Len = 4;

var
  DIBuf: array of Byte;
  L: LongWord;

begin
  try
    Result := ValidateAndEnsureOpen(CardNum);
    if Result <> ERROR_SUCCESS then Exit;

    with Dev[CardNum] do begin

      try
        pTimestamp^ := 0;
      except
        Result := ERROR_NOACCESS;
        Exit;
      end;

      L := EP0Len;
      SetLength(DIBuf, L);
      Result := GenericVendorRead(CardNum, AUR_DIO_WRITE, $0000, 0, L, @DIBuf[0]);
      if (L < EP0Len) or (DIBuf[0] < EP0Len) then begin
        Result := ERROR_INVALID_DATA;
        Exit;
      end;

      pTimestamp^ := PWord(DIBuf[2])^;
      SetLength(DIBuf, 0);

    end;
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;



function WDGInit(CardNum: Integer): LongWord; cdecl;
begin
  Result := ValidateAndEnsureOpen(CardNum);
  if Result <> ERROR_SUCCESS then Exit;

  with Dev[CardNum] do begin

    if not PIDCaps.bCanWDGInit then begin
      Result := ERROR_BAD_TOKEN_TYPE;
      Exit;
    end;

    bWDGInit := True;
  end;
end;

function WDGSetTimeout(CardNum: Integer; var pMilliseconds: Double): LongWord; cdecl;
var
  Secs: Word;

begin
  if pMilliseconds = 0 then
    Secs := 0
  else begin
    Secs := Round(pMilliseconds / 1000);
    if Secs < 1 then Secs := 1;
  end;
  pMilliseconds := Secs * 1000;

  Result := ValidateAndEnsureOpen(CardNum);
  if Result <> ERROR_SUCCESS then Exit;

  with Dev[CardNum] do begin

    if not bWDGInit then begin
      Result := ERROR_BAD_TOKEN_TYPE;
      Exit;
    end;

    Result := GenericVendorWrite(CardNum, AUR_PCIE_WDG, Secs, AUV_TIMEOUT_SET, 0, nil);

  end;
end;

procedure InitConfig(var WDGConfig: TWDGConfig);
begin
  WDGConfig.Length := SizeOf(WDGConfig);
  WDGConfig.Rev := 1;
  WDGConfig.Enabled := True;
  WDGConfig.SecondsBeforeBark := 10;
  WDGConfig.WDG_RESET_ON_BARK := 1;
  WDGConfig.FET_ON_BARK := 1;
  WDGConfig.TTL_ON_BARK := 1;
  WDGConfig.BUZZER_ON_BARK := 0;
  WDGConfig.RELAY_ON_BARK := 1;
  WDGConfig.RED_ON_BARK := $FF;
  WDGConfig.GRN_ON_BARK := $00;
  WDGConfig.BLU_ON_BARK := $00;
  WDGConfig.Null := 0;
end;

function WDGSetResetDuration(CardNum: Integer; var pMilliseconds: Double): LongWord; cdecl;
var
  WDGConfig: TWDGConfig;
  DataSize: LongWord;
  Secs: Byte;

begin
  if pMilliseconds <= 0 then
    Secs := 0
  else if (pMilliseconds <= 1) or (pMilliseconds >= 255000) then
    Secs := 255
  else begin
    Secs := Round(pMilliseconds / 1000);
    if Secs < 1 then Secs := 1;
  end;
  pMilliseconds := Secs * 1000;

  Result := ValidateAndEnsureOpen(CardNum);
  if Result <> ERROR_SUCCESS then Exit;

  with Dev[CardNum] do begin

    if not bWDGInit then begin
      Result := ERROR_BAD_TOKEN_TYPE;
      Exit;
    end;

    InitConfig(WDGConfig);
    DataSize := SizeOf(WDGConfig);
    GenericVendorRead(CardNum, AUR_PCIE_WDG, 0, 0, DataSize, @WDGConfig);

    WDGConfig.WDG_RESET_ON_BARK := Secs;

    Result := GenericVendorWrite(CardNum, AUR_PCIE_WDG, 0, 0, DataSize, @WDGConfig);

  end;
end;

function WDGPet(CardNum: Integer): LongWord; cdecl;
begin
  Result := ValidateAndEnsureOpen(CardNum);
  if Result <> ERROR_SUCCESS then Exit;

  with Dev[CardNum] do begin

    if not bWDGInit then begin
      Result := ERROR_BAD_TOKEN_TYPE;
      Exit;
    end;

    Result := GenericVendorWrite(CardNum, AUR_PCIE_WDG, 0, AUV_PET, 0, nil);

  end;
end;

function CelsiusFromSi7013Temp(Counts: Word): Double;
begin
  Result := 175.72/65536 * Counts - 46.85;
end;

function FahrenheitFromCelsius(Celsius: Double): Double;
begin
  Result := 9/5 * Celsius + 32;
end;

function WDGReadTemp(CardNum: Integer; out TempF: Double): LongWord; cdecl;
var
  DataSize: LongWord;
  WDGStatus: TWDGStatus;

begin
  TempF := 0;

  Result := ValidateAndEnsureOpen(CardNum);
  if Result <> ERROR_SUCCESS then Exit;

  with Dev[CardNum] do begin

    if not bWDGInit then begin
      Result := ERROR_BAD_TOKEN_TYPE;
      Exit;
    end;

    DataSize := SizeOf(WDGStatus);
    Result := GenericVendorRead(CardNum, AUR_PCIE_WDG, 0, 0, DataSize, @WDGStatus);
    if Result <> ERROR_SUCCESS then Exit;

    TempF := FahrenheitFromCelsius(CelsiusFromSi7013Temp(WDGStatus.TempCounts));
  end;
end;

function WDGReadStatus(CardNum: Integer; out StatusByte: LongWord): LongWord; cdecl;
var
  DataSize: LongWord;
  WDGStatus: TWDGStatus;

begin
  StatusByte := $1000;
  Result := ValidateAndEnsureOpen(CardNum);
  if Result <> ERROR_SUCCESS then Exit;

  with Dev[CardNum] do begin

    if not bWDGInit then begin
      Result := ERROR_BAD_TOKEN_TYPE;
      Exit;
    end;

    DataSize := SizeOf(WDGStatus);
    Result := GenericVendorRead(CardNum, AUR_PCIE_WDG, 0, 0, DataSize, @WDGStatus);

    StatusByte := WDGStatus.OldSchoolStatusByte;
    {
      BD0 Watchdog counter refresh reminder (Active low)
      BD1 Temperature good (Active high)
      BD2 Isolated Input #1 status (Same as input)
      BD3 Isolated Input #0 status (Same as input)
      BD4 Fan good (Active high)
      BD5 Power Supply overvoltage (Active low)
      BD6 Power Supply undervoltage (Active low)
      BD7 IRQ generated (Active high)
    }
  end;
end;

function WDGStart(CardNum: Integer): LongWord; cdecl;
var
  WDGConfig: TWDGConfig;
  DataSize: LongWord;

begin
  Result := ValidateAndEnsureOpen(CardNum);
  if Result <> ERROR_SUCCESS then Exit;

  with Dev[CardNum] do begin

    if not bWDGInit then begin
      Result := ERROR_BAD_TOKEN_TYPE;
      Exit;
    end;

    InitConfig(WDGConfig);
    DataSize := SizeOf(WDGConfig);
    GenericVendorRead(CardNum, AUR_PCIE_WDG, 0, 0, DataSize, @WDGConfig);

    GenericVendorWrite(CardNum, AUR_PCIE_WDG, 0, AUV_RESET_WDG, 0, nil);

    WDGConfig.Enabled := True;

    Result := GenericVendorWrite(CardNum, AUR_PCIE_WDG, 0, 0, DataSize, @WDGConfig);
  end;
end;

function WDGStop(CardNum: Integer): LongWord; cdecl;
begin
  Result := ValidateAndEnsureOpen(CardNum);
  if Result <> ERROR_SUCCESS then Exit;

  with Dev[CardNum] do begin

    if not bWDGInit then begin
      Result := ERROR_BAD_TOKEN_TYPE;
      Exit;
    end;

    Result := GenericVendorWrite(CardNum, AUR_PCIE_WDG, 0, AUV_DISABLE, 0, nil);

  end;
end;

procedure TBulkPseudoIRQThread.Execute;
var
  Flags, ShutdownFlag: set of (ewxShutdown, ewxReboot, ewxForce, ewxPoweroff, ewxForceIfHung);
  Disable: Boolean;
  Timeout: Double;
  Status: LongWord;

begin
  ReturnValue := ERROR_CANCELLED;
  Status := WaitForWDG(CardNum);
  //if MessageBoxW(0, PWideChar('Error ' + IntToStr(Status)), 'WaitForIRQ Return', MB_OKCANCEL) = IDCANCEL then Exit;
  if Status <> ERROR_SUCCESS then Exit;

  //If we get to here, an IRQ has just occurred
  //if MessageBoxW(0, PWideChar('WDGAction ' + IntToStr(Dev[CardNum].WDGAction)), 'WDGAction', MB_OKCANCEL) = IDCANCEL then Exit;
  Flags := [];
  Disable := (Dev[CardNum].WDGAction and (1 shl 0)) <> 0;
  if Disable or bStupidXPShutdownIssue then
    ShutdownFlag := [ewxReboot]
  else
    ShutdownFlag := [ewxShutdown]
  ;
  if (Dev[CardNum].WDGAction and (1 shl 1)) <> 0 then Flags := Flags + ShutdownFlag;
  if (Dev[CardNum].WDGAction and (1 shl 2)) <> 0 then Flags := Flags + ShutdownFlag + [ewxForce];

  if Disable then begin
    WDGStop(CardNum);
  end
  else if Flags <> [] then begin
    Timeout := 90000;
    WDGSetTimeout(CardNum, Timeout);
  end;
  if Flags <> [] then begin
    if (OS = osNT) or (OS = osNT5) then GetPrivilege('SeShutdownPrivilege');
    ExitWindowsEx(Byte(Flags), 0);
  end;
end;

function WDGHandleIRQ(CardNum: Integer; Action: LongWord): LongWord; cdecl;
begin
  if Action > 5 then begin
    Result := ERROR_INVALID_PARAMETER;
    Exit;
  end;

  Result := ValidateAndEnsureOpen(CardNum);
  if Result <> ERROR_SUCCESS then Exit;

  with Dev[CardNum] do begin

    if not bWDGInit then begin
      Result := ERROR_BAD_TOKEN_TYPE;
      Exit;
    end;

    Result := GenericVendorWrite(CardNum, AUR_PCIE_WDG, 0, AUV_NOTIRQ_CLEAR, 0, nil);
    if Result <> ERROR_SUCCESS then Exit;

    Dev[CardNum].WDGAction := Action;
    BulkPseudoIRQThread := TBulkPseudoIRQThread.Create(True);
    BulkPseudoIRQThread.CardNum := CardNum;
    BulkPseudoIRQThread.Start;
  end;
end;

function EmergencyReboot(): LongWord; cdecl;
begin
  if (OS = osNT) or (OS = osNT5) then GetPrivilege('SeShutdownPrivilege');
  if ExitWindowsEx(EWX_REBOOT or EWX_FORCE, 0) then
    Result := ERROR_SUCCESS
  else begin
    Result := GetLastError;
    if Result = ERROR_SUCCESS then Result := ERROR_GEN_FAILURE;
  end;
end;





function _DI_ReadAll(Value: Word; CardNum: Integer; pDIData: PByte): LongWord;
const
  EP0Len = 7;

var
  DIBuf: array of Byte;
  L: LongWord;

begin
  try
    Result := ValidateAndEnsureOpen(CardNum);
    if Result <> ERROR_SUCCESS then Exit;

    with Dev[CardNum] do begin

      try
        pDIData^ := 0;
      except
        Result := ERROR_NOACCESS;
        Exit;
      end;

      L := EP0Len;
      SetLength(DIBuf, L);
      Result := GenericVendorRead(CardNum, AUR_DIO_WRITE, Value, 0, L, @DIBuf[0]);
      if (L < EP0Len) or (DIBuf[0] < EP0Len) then begin
        Result := ERROR_INVALID_DATA;
        Exit;
      end;

      {
      Buf := '';
      for iByte := 0 to High(DIBuf) do
        Buf := Buf + ('[' + IntToHex(iByte, 2) + '] ' + IntToHex(DIBuf[iByte], 2) + #13#10)
      ;
      Clipboard.AsText := Buf;
      }

      pDIData^ := DIBuf[6];
      SetLength(DIBuf, 0);

    end;
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;

function _DI_Read1(Value: Word; CardNum: Integer; BitIndex: Byte; pBitValue: PLongBool): LongWord;
var
  DIData: Byte;

begin
  try

    try
      pBitValue^ := False;
    except
      Result := ERROR_NOACCESS;
      Exit;
    end;

    if BitIndex >= 8 then begin
      Result := ERROR_INVALID_INDEX;
      Exit;
    end;

    Result := _DI_ReadAll(Value, CardNum, @DIData);
    if Result = ERROR_SUCCESS then
      pBitValue^ := (DIData and (1 shl BitIndex)) <> 0
    ;

  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;

function DI_ReadAll(CardNum: Integer; pDIData: PByte): LongWord; cdecl;
begin
  Result := _DI_ReadAll($0001, CardNum, pDIData);
end;

function DI_Read1(CardNum: Integer; BitIndex: Byte; pBitValue: PLongBool): LongWord; cdecl;
begin
  Result := _DI_Read1($0001, CardNum, BitIndex, pBitValue);
end;

function DI_ReadLatched(CardNum: Integer; pDIRisen, pDIFallen: PByte): LongWord; cdecl;
const
  EP0Len = 9;

var
  DIBuf: array of Byte;
  L: LongWord;

begin
  try
    Result := ValidateAndEnsureOpen(CardNum);
    if Result <> ERROR_SUCCESS then Exit;

    with Dev[CardNum] do begin

      try
        pDIRisen^ := 0;
        pDIFallen^ := 0;
      except
        Result := ERROR_NOACCESS;
        Exit;
      end;

      L := EP0Len;
      SetLength(DIBuf, L);
      Result := GenericVendorRead(CardNum, AUR_DIO_WRITE, $0001, 0, L, @DIBuf[0]);
      if (L < EP0Len) or (DIBuf[0] < EP0Len) then begin
        Result := ERROR_INVALID_DATA;
        Exit;
      end;

      pDIRisen^ := DIBuf[7];
      pDIFallen^ := DIBuf[8];
      SetLength(DIBuf, 0);

    end;
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;

function DI_PeekAll(CardNum: Integer; pDIData: PByte): LongWord; cdecl;
begin
  Result := _DI_ReadAll($0000, CardNum, pDIData);
end;

function DI_Peek1(CardNum: Integer; BitIndex: Byte; pBitValue: PLongBool): LongWord; cdecl;
begin
  Result := _DI_Read1($0000, CardNum, BitIndex, pBitValue);
end;

function DI_SetDebounceAll(CardNum: Integer; TimeMS: Byte): LongWord; cdecl;
begin
  try
    Result := ValidateAndEnsureOpen(CardNum);
    if Result <> ERROR_SUCCESS then Exit;

    with Dev[CardNum] do begin

      Result := GenericVendorWrite(CardNum, AUR_DIO_WRITE, $FF00 or TimeMS, $FFFF, 0, nil);

    end;
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;

function DI_SetDebounce1(CardNum: Integer; BitIndex: Byte; TimeMS: Byte): LongWord; cdecl;
begin
  try
    Result := ValidateAndEnsureOpen(CardNum);
    if Result <> ERROR_SUCCESS then Exit;

    with Dev[CardNum] do begin

      Result := GenericVendorWrite(CardNum, AUR_DIO_WRITE, ($100 shl BitIndex) or TimeMS, $FFFF, 0, nil);

    end;
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;

function DI_SetDebounceSome(CardNum: Integer; BitMask: Byte; TimeMS: Byte): LongWord; cdecl;
begin
  try
    Result := ValidateAndEnsureOpen(CardNum);
    if Result <> ERROR_SUCCESS then Exit;

    with Dev[CardNum] do begin

      Result := GenericVendorWrite(CardNum, AUR_DIO_WRITE, (BitMask shl 8) or TimeMS, $FFFF, 0, nil);

    end;
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;

function DO_WriteAll(CardNum: Integer; DOData: Byte): LongWord; cdecl;
begin
  try
    Result := ValidateAndEnsureOpen(CardNum);
    if Result <> ERROR_SUCCESS then Exit;

    with Dev[CardNum] do begin

      Result := GenericVendorWrite(CardNum, AUR_DIO_WRITE, DOData, 0, 0, nil);

    end;
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;

function DO_Write1(CardNum: Integer; BitIndex: Byte; BitValue: LongBool): LongWord; cdecl;
var
  V: Word;

begin
  try
    Result := ValidateAndEnsureOpen(CardNum);
    if Result <> ERROR_SUCCESS then Exit;

    with Dev[CardNum] do begin

      if BitIndex >= 8 then begin
        Result := ERROR_INVALID_INDEX;
        Exit;
      end;

      V := BitIndex;
      V := V shl 8;
      if BitValue then V := V or $01;
      Result := GenericVendorWrite(CardNum, AUR_DIO_WRITE, V, 1, 0, nil);

    end;
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;





procedure UnpackAICounts(pAIData: PDouble; pCtSrc: PWord);
begin
  pAIData^ := Si7013_TempToCelsius(pCtSrc^);
  Inc(pAIData); Inc(pCtSrc);
  pAIData^ := Si7013_HumidityToRH(pCtSrc^);
  Inc(pAIData); Inc(pCtSrc);
  pAIData^ := Si7013_CountsToMilliVolts(pCtSrc^);
  Inc(pAIData); Inc(pCtSrc);
  pAIData^ := Si7013_CountsToMilliVolts(pCtSrc^);
  Inc(pAIData); Inc(pCtSrc);
  pAIData^ := Si7013_CountsToMilliVolts(pCtSrc^);
  Inc(pAIData); Inc(pCtSrc);
  pAIData^ := Si7013_CountsToMilliVolts(pCtSrc^);
  //Inc(pAIData); Inc(pCtSrc);
end;

function AI_ReadAll(CardNum: Integer; pAIData: PDouble): LongWord; cdecl;
const
  EP0Len = 9+12;

var
  AIBuf: array of Byte;
  L: LongWord;

begin
  try
    Result := ValidateAndEnsureOpen(CardNum);
    if Result <> ERROR_SUCCESS then Exit;

    with Dev[CardNum] do begin

      try
        FillChar(pAIData^, SizeOf(pAIData^) * 6, 0);
      except
        Result := ERROR_NOACCESS;
        Exit;
      end;

      if not bAIEnabled then begin
        L := 0;
        bAIEnabled := GenericVendorWrite(CardNum, $40, $1, $0001, L, nil) = ERROR_SUCCESS;
      end;

      L := EP0Len;
      SetLength(AIBuf, L);
      Result := GenericVendorRead(CardNum, AUR_DIO_WRITE, $0000, 0, L, @AIBuf[0]);
      if (L < EP0Len) or (AIBuf[0] < EP0Len) then begin
        Result := ERROR_INVALID_DATA;
        Exit;
      end;

      UnpackAICounts(pAIData, @AIBuf[9]);
      SetLength(AIBuf, 0);
    end;
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;

function AI_Read1(CardNum: Integer; ChannelIndex: Byte; pAIData: PDouble): LongWord; cdecl;
var
  AIData: array of Double;

begin
  try
    try
      pAIData^ := 0;
    except
      Result := ERROR_NOACCESS;
      Exit;
    end;

    if ChannelIndex >= 6 then begin
      Result := ERROR_INVALID_INDEX;
      Exit;
    end;

    SetLength(AIData, 6);
    Result := AI_ReadAll(CardNum, @AIData[0]);
    if Result <> ERROR_SUCCESS then Exit;

    pAIData^ := AIData[ChannelIndex];
    SetLength(AIData, 0);
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;



function InPortB(Addr: LongWord): Word; cdecl;
var
  CardNum: Integer;

begin
  CardNum := Addr div 16;
  if CardNum >= 32 then CardNum := 0;
  Result := RelInPortB(CardNum, Addr mod 16);
end;

function OutPortB(Addr: LongWord; Value: Byte): Word; cdecl;
var
  CardNum: Integer;

begin
  CardNum := Addr div 16;
  if CardNum >= 32 then CardNum := 0;
  Result := RelOutPortB(CardNum, Addr mod 16, Value);
end;

function RelInPortB(CardNum, Offset: LongWord): Word; cdecl;
var
  Buf32: LongWord;
  BufR: Double;
  Status: LongWord;

begin
  Result := $FF;
  Status := ERROR_SUCCESS;
  case Offset of
    $0: ;
    $1: ;
    $2: ;
    $3: ;
    $4: begin Status := WDGReadStatus(CardNum, Buf32); Result := Buf32; end;
    $5: begin Status := WDGReadTemp(CardNum, BufR); Result := Round(15/11 * (BufR - 7)); end;
    $6: ;
    $7: Status := WDGStop(CardNum);
    $8: begin Status := WDGReadStatus(CardNum, Buf32); Result := Buf32; end;
    $9: begin Status := WDGReadStatus(CardNum, Buf32); Result := Buf32; end;
    $A: ;
    $B: ;
    $C: ;
    $D: ;
    $E: ;
    $F: ;
  end;
  if Status <> ERROR_SUCCESS then begin
    SetLastError(Status);
    Result := $AA55;
  end;
end;

function RelOutPortB(CardNum, Offset: LongWord; Value: Byte): Word; cdecl;
var
  Status: LongWord;

begin
  Result := 0;
  Status := ERROR_SUCCESS;
  case Offset of
    $0: ;
    $1: ;
    $2: Status := WDGPet(CardNum);
    $3: ;
    $4: ;
    $5: ;
    $6: ;
    $7: Status := WDGStart(CardNum);
    $8: ;
    $9: ;
    $A: ;
    $B: ;
    $C: ;
    $D: ;
    $E: ;
    $F: ;
  end;
  if Status <> ERROR_SUCCESS then begin
    SetLastError(Status);
    Result := $AA55;
  end;
end;





function AIOWDG_ResetChip(CardNum: Integer): LongWord; cdecl;
var
  CPUCSByte: Byte;

begin
  try
    //Put CPU into reset
    CPUCSByte := $01;
    Result := GenericVendorWrite(CardNum, $A0, $E600, 0, 1, @CPUCSByte);
    if Result <> ERROR_SUCCESS then Exit;

    //take CPU out of reset
    CPUCSByte := $00;
    Result := GenericVendorWrite(CardNum, $A0, $E600, 0, 1, @CPUCSByte);
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;





function GetOS: TOS;
var
  V: OSVERSIONINFO;
  VerNum: LongWord;
  VerNums: packed record Min, Maj: Word; end absolute VerNum;

begin
  Result := osDunno; //Assume we're on Win9x if we can't tell where we are

  V.dwOSVersionInfoSize := sizeof(OSVERSIONINFO);
  GetVersionEx(V);
  VerNums.Min := V.dwMinorVersion;
  VerNums.Maj := V.dwMajorVersion;
  case V.dwPlatformId of
  VER_PLATFORM_WIN32s: Result := osWin95; // Win32s doesn't have WDM, does it?
  VER_PLATFORM_WIN32_WINDOWS:
    if VerNum < $4000A then
      Result := osWin95
    else
      Result := osWin98
    ;
  VER_PLATFORM_WIN32_NT:
    if VerNum < $50000 then
      Result := osNT
    else begin
      Result := osNT5;
      if VerNum >= $50001 then
        bStupidXPShutdownIssue := True
      ;
    end;
  end;
end;

begin
  if OS = osDunno then OS := GetOS;
end.
